@@ -19,11 +19,11 @@
 import sys
 import os
 
-from samba.tests.krb5.kdc_base_test import KDCBaseTest
-
 sys.path.insert(0, "bin/python")
 os.environ["PYTHONUNBUFFERED"] = "1"
 
+from samba.tests.krb5.kdc_base_test import KDCBaseTest
+
 global_asn1_print = False
 global_hexdump = False
 
@@ -65,7 +65,9 @@ class RodcKerberosTests(KDCBaseTest):
                                                  to_rodc=True)
 
         # Ensure the PAC contains the expected checksums.
-        self.verify_ticket(service_ticket, rodc_key, service_ticket=True)
+        self.verify_ticket(service_ticket, rodc_key, service_ticket=True,
+                           expect_ticket_checksum=True,
+                           expect_full_checksum=True)
 
 
 if __name__ == "__main__":
